/*
** Hash-A-Lot by Ben Slusky <sluskyb@paranoiacs.org>
** 
** This program will read a passphrase from standard input and print a binary
** (not printable) hash to standard output. The output is suitable for use as
** an encryption key.
**
** USAGE:
**		hashalot [ -q ] [ -x ] [ -s _salt_ ] [ -n _#bytes_ ] _hashtype_
**	OR
**		_hashtype_ [ -q ] [ -x ] [ -s _salt_ ] [ -n _#bytes_ ]
**
** Most of the code was cribbed from the kerneli.org patch to util-linux,
** by Marc Mutz <Marc@Mutz.com>. Most of what wasn't, was cribbed from GnuPG,
** v.1.0.3 (http://www.gnupg.org/).
*/

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <error.h>

#include <sys/types.h>
#include <sys/mman.h>

#include "rmd160.h"
#include "sha512.h"
#include "getpass.h"

#define PASSWDBUFFLEN 130

const char *progname;

typedef int (*phash_func_t) (char *dest, size_t dest_len,
			     const char *src, size_t src_len);

static int
phash_rmd160(char *dest, size_t dest_len, const char *src, size_t src_len)
{
	char    tmp[PASSWDBUFFLEN] = { 'A', 0, };
	char    key[RMD160_HASH_SIZE * 2] = { 0, };

	strncat(tmp, src, sizeof(tmp) - 1);

	rmd160_hash_buffer(key, src, src_len);
	rmd160_hash_buffer(key + RMD160_HASH_SIZE, tmp,
			   src_len + 1 /* dangerous! */ );

	memcpy(dest, key, dest_len);

	memset(tmp, 0, sizeof tmp);	/* paranoia */
	memset(key, 0, sizeof key);	/* paranoia */

	return dest_len;
}

static int
phash_sha256(char *dest, size_t dest_len, const char *src, size_t src_len)
{
	memset(dest, 0, dest_len);
	sha256_hash_buffer((const unsigned char *) src, src_len,
			   (unsigned char *) dest, dest_len);
	return dest_len;
}

static int
phash_sha384(char *dest, size_t dest_len, const char *src, size_t src_len)
{
	memset(dest, 0, dest_len);
	sha384_hash_buffer((const unsigned char *) src, src_len,
			   (unsigned char *) dest, dest_len);
	return dest_len;
}

static int
phash_sha512(char *dest, size_t dest_len, const char *src, size_t src_len)
{
	memset(dest, 0, dest_len);
	sha512_hash_buffer((const unsigned char *) src, src_len,
			   (unsigned char *) dest, dest_len);
	return dest_len;
}

struct func_table_t
{
	const char *name;
	phash_func_t func;
	size_t def_length;
} static func_table[] = {
	{ "ripemd160",	phash_rmd160, 20 },
	{ "rmd160",	phash_rmd160, 20 },
	{ "rmd160compat", phash_rmd160, 16 },
	{ "sha256",	phash_sha256, 32 },
	{ "sha384",	phash_sha384, 48 },
	{ "sha512",	phash_sha512, 64 },
	{ 0, 0, 0 }
};

static int __attribute__ ((__noreturn__))
show_usage(void)
{
	struct func_table_t *p = func_table;

	fprintf(stderr,
		"usage:\n"
		"    hashalot [ -q ] [ -x ] [ -s SALT ] [ -n _#bytes_ ] HASHTYPE\n"
		"  or\n"
		"    HASHTYPE [ -q ] [ -x ] [ -s SALT ] [ -n _#bytes_ ]\n"
		"\n" "supported values for HASHTYPE:");

	for (; p->name; ++p)
		fprintf(stderr, " %s", p->name);

	fprintf(stderr, "\n");

	exit(EXIT_FAILURE);
}

static  phash_func_t
phash_lookup(const char *phash_name, size_t *length)
{
	struct func_table_t *p = func_table;

	if (!phash_name)
		return 0;

	for (; p->name && strcmp(phash_name, p->name); ++p)
		;
	if (length)
		*length = p->def_length;
	return p->func;
}

static char *
zeroize(char *str)
{
	char *p = str;

	while (*p)
		*p++ = '\0';
	return str;
}

/* function to append a "salt" to the passphrase, to better resist
 * dictionary attacks */
static char *
salt_passphrase(char *pass, const char *salt)
{
	char   *buf;
	int     rc;

	rc = asprintf(&buf, "%s%s", pass, salt);
	zeroize(pass);
	if (rc < 0)
		error(EXIT_FAILURE, errno, "asprintf");

	free(pass);
	return buf;
}

static void *
xmalloc(size_t size)
{
	void   *r = malloc(size);

	if (!r)
		error(EXIT_FAILURE, errno, "malloc");
	return r;
}

static void
hexify(char *hash, size_t hashlen)
{
	int     i;
	char   *h = xmalloc(hashlen);

	memcpy(h, hash, hashlen);
	for (i = 0; i < hashlen; i++)
		snprintf((hash + 2 * i), 3, "%.2x", (unsigned char) h[i]);
	strcat(hash, "\n");

	memset(h, 0, hashlen);	/* paranoia */
	free(h);
}

static int
write_loop(int fd, const char *buffer, size_t count)
{
	ssize_t offset = 0;

	while (count > 0)
	{
		ssize_t block = write(fd, &buffer[offset], count);

		if (block < 0)
		{
			if (errno == EINTR)
				continue;
			return block;
		}
		if (!block)
			break;
		offset += block;
		count -= block;
	}
	return offset;
}

int
main(int argc, char *argv[])
{
	char   *pass, *passhash, *salt = NULL, *p;
	size_t  hashlen = 0;
	phash_func_t func;
	int     quiet = 0, hex_output = 0, c;
	int     rc = EXIT_SUCCESS;

	progname = (argc > 0) ? basename(argv[0]) : "hashalot";

	while ((c = getopt(argc, argv, "n:qs:x")) != -1)
	{
		switch (c)
		{
			case 'n':
				hashlen = strtoul(optarg, &p, 0);
				if (*p != '\0' || *optarg == '\0')
				{
					fprintf(stderr,
						"%s: argument to -n option must be numeric\n",
						progname);
					show_usage();
					exit(EXIT_FAILURE);
				}
				break;
			case 'q':
				quiet = 1;
				break;
			case 's':
				salt = optarg;
				break;
			case 'x':
				hex_output = 1;
				break;
			default:
				return EXIT_FAILURE;
		}
	}

	if (!(func = phash_lookup(progname, (hashlen ? NULL : &hashlen))))
	{
		/* lookup failed, so try next argv */
		if (optind + 1 != argc)
			show_usage();
		func = phash_lookup(argv[optind],
				    (hashlen ? NULL : &hashlen));
		if (!func)
		{
			/* lookup failed again */
			fprintf(stderr,
				"%s: missing or unknown hash type requested\n",
				progname);
			show_usage();
		}
	}

	/* allocate memory for the password hash:
	 * enough for 2 hex digits per byte of the requested hash length,
	 * plus a newline, plus a null */
	passhash = xmalloc(2 * hashlen + 2);

	/* try to lock memory so it doesn't get swapped out for sure */
	if (mlockall(MCL_CURRENT | MCL_FUTURE) && !quiet)
		error(EXIT_SUCCESS, errno, "failed to lock memory");

	/* here we acquire the precious passphrase... */
	if (!(pass = xgetpass("Enter passphrase: ")))
		error(EXIT_FAILURE, errno, "getpass failed");
	if (salt)
		pass = salt_passphrase(pass, salt);
	hashlen = func(passhash, hashlen, pass, strlen(pass));
	free(zeroize(pass));

	if (hex_output)
	{
		hexify(passhash, hashlen);
		hashlen = hashlen * 2 + 1;
	}

	if (write_loop(STDOUT_FILENO, passhash, hashlen) != hashlen)
	{
		error(EXIT_SUCCESS, errno, "write");
		rc = EXIT_FAILURE;
	}

	memset(passhash, 0, hashlen);	/* paranoia again */
	free(passhash);

	return rc;
}
